import numpy as np


def neg_log_like(beta, game_matrix_list):
    '''
    compute the negative loglikelihood
    ------------
    Input:

    beta: can be a T-by-N array or a T*N-by-1 array
    game_matrix_list: records of games, T-by-N-by-N array
    ------------
    -l: negative loglikelihood, a number
    '''
    # beta could be a T-by-N matrix or T*N-by-1 array
    T, N = game_matrix_list.shape[0:2]
    beta = beta.reshape(T, N)
    # l stores the loglikelihood
    l = 0
    N_one = np.ones(N).reshape(N, 1)
    for t in range(T):
        Cu = np.triu(game_matrix_list[t])  # equivalent to [t,:,:]
        Cl = np.tril(game_matrix_list[t])
        b = beta[t, :].reshape(N, 1)
        D = b @ N_one.T - N_one @ b.T
        W = np.log(1 + np.exp(D))
        l += N_one.T @ (Cu * D) @ N_one - \
            N_one.T @ ((Cu + Cl.T) * np.triu(W)) @ N_one
    return -l[0, 0]


def grad_nl(beta, game_matrix_list):
    '''
    compute the gradient of the negative loglikelihood
    ------------
    Input:
    beta: can be a T-by-N array or a T*N-by-1 array
    game_matrix_list: records of games, T-by-N-by-N array
    ------------
    Output:
    -grad: gradient of negative loglikelihood, a T*N-by-1 array
    '''
    # beta could be a T-by-N array or a T*N-by-1 array
    T, N = game_matrix_list.shape[0:2]
    beta = beta.reshape(T, N)
    # g stores the gradient
    g = np.zeros(N * T).reshape(T, N)
    N_one = np.ones(N).reshape(N, 1)
    for t in range(T):
        C = game_matrix_list[t]
        b = beta[t, :].reshape(N, 1)
        W = np.exp(b @ N_one.T) + np.exp(N_one @ b.T)
        g[t, :] = ((C / W) @ np.exp(b) - (C / W).T @ N_one * np.exp(b)).ravel()
    return -g.reshape(N * T, 1)


def hess_nl(beta, game_matrix_list):
    '''
    compute the Hessian of the negative loglikelihood
    ------------
    Input:
    beta: can be a T-by-N array or a T*N-by-1 array
    game_matrix_list: records of games, T-by-N-by-N array
    ------------
    Output:
    -H: Hessian of negative loglikelihood T*N-by-T*N array
    '''
    # beta could be a T-by-N array or a T*N-by-1 array
    T, N = game_matrix_list.shape[0:2]
    beta = beta.reshape(T, N)
    # H stores the Hessian
    H = np.zeros(N ** 2 * T ** 2).reshape(T * N, T * N)
    N_one = np.ones(N).reshape(N, 1)
    for t in range(T):
        Cu = np.triu(game_matrix_list[t])  # equivalent to [t,:,:]
        Cl = np.tril(game_matrix_list[t])
        Tm = Cu + Cl.T + Cu.T + Cl
        b = beta[t, :].reshape(N, 1)
        W = np.exp(b @ N_one.T) + np.exp(N_one @ b.T)
        H_t = Tm * np.exp(b @ N_one.T + N_one @ b.T) / W ** 2
        H_t += -np.diag(sum(H_t))
        ind = range(t * N, (t + 1) * N)
        H[t * N:(t + 1) * N, t * N:(t + 1) * N] = H_t
    return -H


'''
##### example

N = 10
T = 10
tn_median = 100
bound_games = [tn_median - 2,tn_median + 2] # bounds for the number of games between each pair of teams

##### tn: list of number of games between each pair of teams
tn = stats.randint.rvs(low = int(bound_games[0]), high = int(bound_games[1]), size = int(T * N * (N - 1) / 2))

beta0 = beta_gaussian_process(N, T, mu_parameters = [0,1], cov_parameters = [alpha,r], mu_type = 'constant', cov_type = 'toeplitz')
game_matrix_list = get_game_matrix_list(N,T,tn,beta0)


loglike(beta0,game_matrix_list)
g = grad_l(beta0,game_matrix_list)
H = Hess_l(beta0,game_matrix_list)





##### test case (numerically)

# verify gradient
delta = 0.0001
g_hat = np.zeros((N * T,1))

for i in range(N * T):
    beta1 = beta0.copy().reshape(N * T,1).ravel()
    beta2 = beta1.copy()
    beta1[i] -= delta
    beta2[i] += delta
    g_hat[i] = (loglike(beta2,game_matrix_list) - loglike(beta1,game_matrix_list)) / (2 * delta)

print(sum(abs(g - g_hat)))


# verify Hessian
delta = 0.0001
H_hat = np.zeros((N * T,N * T))

for i in range(N * T):
    beta1 = beta0.copy().reshape(N * T,1).ravel()
    beta2 = beta1.copy()
    beta1[i] -= delta
    beta2[i] += delta
    H_hat[i] = (grad_l(beta2,game_matrix_list) - grad_l(beta1,game_matrix_list)).ravel() / (2 * delta)

print(sum(sum(abs(H_hat - H))))

'''



import scipy as sc
from scipy import stats
import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse import diags


def beta_gaussian_process(
        N,
        T,
        mu_parameters,
        cov_parameters,
        mu_type='constant',
        cov_type='toeplitz'):
    '''
    generate beta via a Gaussian process
    '''
    if mu_type == 'constant':
        loc, scale = mu_parameters
        mu_start = stats.norm.rvs(
            loc=loc, scale=scale, size=N, random_state=100)
        mu = [np.ones(T) * mu_start[i] for i in range(N)]
    if cov_type == 'toeplitz':
        alpha, r = cov_parameters
    # strong auto-correlation case, off diagonal  = 1 - T^(-alpha) * |i - j|^r
    off_diag = 1 - T ** (-alpha) * np.arange(1, T + 1) ** r
    cov_single_path = sc.linalg.toeplitz(off_diag, off_diag)

    return np.array([np.random.multivariate_normal(
        mean=mu[i], cov=cov_single_path, size=1).ravel() for i in range(N)]).T


def get_game_matrix_list(N, T, tn, beta):
    '''
    get the list of T game matrices
    -------------
    Input:
    N: number of teams
    T: number of seasons
    tn: list of number of games between each pair of teams
    beta: a T-by-N array
    -------------
    Output:
    game_matrix_list: a 3-d np.array of T game matrices, each matrix (t,:,:) stores the number of times that i wins j at entry (i,j) at season t
    '''
    game_matrix_list = [None] * T
    ind = -1  # a counter used to get tn
    for t in range(T):
        game_matrix = np.zeros(N * N).reshape(N, N)
        for i in range(N):
            for j in range(i + 1, N):
                ind += 1
                pij = np.exp(beta[t, i] - beta[t, j]) / \
                    (1 + np.exp(beta[t, i] - beta[t, j]))
                nij = np.random.binomial(n=tn[ind], p=pij, size=1)
                game_matrix[i, j], game_matrix[j, i] = nij, tn[ind] - nij
        game_matrix_list[t] = game_matrix
    return np.array(game_matrix_list)


def beta_markov_chain(
        num_team,
        num_season,
        var_latent=1,
        coef_latent=1,
        sig_latent=1,
        draw=True):
    pc_latent = diags([-coef_latent / var_latent, 1 / var_latent, -coef_latent / var_latent],
                      offsets=[-1, 0, 1],
                      shape=(num_season, num_season)).todense()
    pc_latent[np.arange(num_season - 1), np.arange(num_season - 1)
              ] += (coef_latent**2) / var_latent
    var_latent = np.linalg.inv(pc_latent)
    inv_sqrt_var = np.diag(1 / np.sqrt(np.diag(var_latent)))
    var_latent = sig_latent * inv_sqrt_var @ var_latent @ inv_sqrt_var

    latent = np.transpose(
        np.random.multivariate_normal(
            [0] * num_season, var_latent, num_team))

    if draw:
        f = plt.figure(1, figsize=(12, 5))

        ax = plt.subplot(121)
        plt.imshow(var_latent,
                   cmap='RdBu', vmin=-sig_latent, vmax=sig_latent)
        plt.colorbar()
        plt.title("variance")

        ax = plt.subplot(122)
        plt.imshow(latent, cmap='RdBu',
                   vmin=-np.max(np.abs(latent)),
                   vmax=np.max(np.abs(latent)))
        plt.xlabel("team number")
        plt.ylabel("season number")
        plt.title("beta")
    return latent


'''
some examles of running functions beta_gaussian_process and get_game_matrix_list
##### example of generating

N = 10 # number of teams
T = 10 # number of seasons/rounds/years
tn_median = 100
bound_games = [tn_median - 2,tn_median + 2] # bounds for the number of games between each pair of teams

##### tn: list of number of games between each pair of teams
tn = stats.randint.rvs(low = int(bound_games[0]), high = int(bound_games[1]), size = int(T * N * (N - 1) / 2))

##### get beta here #####
beta = beta_gaussian_process(N, T, mu_parameters = [0,1], cov_parameters = [alpha,r], mu_type = 'constant', cov_type = 'toeplitz')

game_matrix_list = get_game_matrix_list(N,T,tn,beta)


##### example of drawing paths

beta = beta_gaussian_process(N, T, mu_parameters = [beta_mu], cov_parameters = [alpha,r], mu_type = 'constant', cov_type = 'toeplitz')

f = plt.figure(3, figsize = (9,4.5))
ax = plt.subplot(111)
for i in range(T):
    ax.plot(range(1,T + 1),beta[i],c=np.random.rand(3,1),marker = '.',label = 'Team' + str(i),linewidth=1)

box = ax.get_position()
ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

# Put a legend to the right of the current axis
ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
plt.show()
'''



#!/usr/bin/env python
import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()
doclink = """
Documentation
-------------

The full documentation is at http://bttv.rtfd.org."""
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='bttv',
    version='0.1.0',
    description='Tools for a implementing time-varying Bradley-Terry rankings',
    long_description=readme + '\n\n' + doclink + '\n\n' + history,
    author="Heejong Bong <blueconet@gmail.com>, Wanshan Li <wanshanl@andrew.cmu.edu>, Shamindra Shrotriya <shamindra.shrotriya@gmail.com>",
    author_email='shamindra.shrotriya@gmail.com',
    url='https://github.com/shamindras/bttv',
    packages=[
        'bttv',
    ],
    package_dir={'bttv': 'bttv'},
    include_package_data=True,
    install_requires=[
    ],
    license='MIT',
    zip_safe=False,
    keywords='bttv',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: PyPy',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
    ],
    python_requires='>=3.6',
)



import sys
import numpy as np
import scipy as sc
import scipy.linalg as spl


def loocv(data, lambdas_smooth, opt_fn,
          num_loocv=200, get_estimate=True,
          verbose='cv', out='terminal', **kwargs):
    '''
    conduct local
    ----------
    Input:
    data: TxNxN array
    lambdas_smooth: a vector of query lambdas
    opt_fn: a python function in a particular form of
        opt_fn(data, lambda_smooth, beta_init=None, **kwargs)
        kwargs might contain hyperparameters
        (e.g., step size, max iteration, etc.) for
        the optimization function
    num_loocv: the number of random samples left-one-out cv sample
    get_estimate: whether or not we calculate estimates beta's for
        every lambdas_smooth. If True, we use those estimates as
        initial values for optimizations with cv data
    verbose: controlling the verbose level. If 'cv', the function
        prints only cv related message. If 'all', the function prints
        all messages including ones from optimization process.
        The default is 'cv'.
    out: controlling the direction of output. If 'terminal', the function
        prints into the terminal. If 'notebook', the function prints into
        the ipython notebook. If 'file', the function prints into a log
        file 'cv_log.txt' at the same directory. You can give a custom
        output stream to this argument. The default is 'terminal'
    **kwargs: keyword arguments for opt_fn
    ----------
    Output:
    lambda_cv: lambda_smooth chosen after cross-validation
    nll_cv: average cross-validated negative loglikelihood
    beta_cv: beta chosen after cross-validation. None if get_estimate is False
    '''
    lambdas_smooth = lambdas_smooth.flatten()
    lambdas_smooth = -np.sort(-lambdas_smooth)
    betas = [None] * lambdas_smooth.shape[0]

    last_beta = np.zeros(data.shape[:2])
    for i, lambda_smooth in enumerate(lambdas_smooth):
        _, beta = opt_fn(data, lambda_smooth, beta_init=last_beta, **kwargs)
        betas[i] = beta.reshape(data.shape[:2])
        last_beta = betas[i]

    indices = np.array(np.where(np.full(data.shape, True))).T
    cum_match = np.cumsum(data.flatten())

    if out == 'terminal':
        out = sys.__stdout__
    elif out == 'notebook':
        out = sys.stdout
    elif out == 'file':
        out = open('cv_log.txt', 'w')

    loglikes_loocv = np.zeros(lambdas_smooth.shape)
    for i in range(num_loocv):
        data_loocv = data.copy()
        rand_match = np.random.randint(np.sum(data))
        rand_index = indices[np.min(np.where(cum_match >= rand_match)[0])]
        data_loocv[tuple(rand_index)] -= 1

        for j, lambda_smooth in enumerate(lambdas_smooth):
            _, beta_loocv = opt_fn(
                data_loocv, lambda_smooth, beta_init=betas[j], verbose=(
                    verbose in ['all']), out=out, **kwargs)
            beta_loocv = beta_loocv.reshape(data.shape[:2])
            loglikes_loocv[j] += beta_loocv[rand_index[0], rand_index[1]] \
                - np.log(np.exp(beta_loocv[rand_index[0], rand_index[1]])
                         + np.exp(beta_loocv[rand_index[0], rand_index[2]]))

        if verbose in ['cv', 'all']:
            out.write("%d-th cv done\n" % (i + 1))
            out.flush()

    return (lambdas_smooth[np.argmax(loglikes_loocv)], -
            loglikes_loocv[::-
                           1] /
            num_loocv, betas[np.argmax(loglikes_loocv)])



import sys
import numpy as np
import scipy as sc
import scipy.linalg as spl
import bttv.grad_utils as model


########################## squared l2 penalty ############################


def objective_l2_sq(beta, game_matrix_list, l_penalty):
    '''
    compute the objective of the model (neg_log_like + l2_square)
    ----------
    Input:
    beta: TxN array or a TN vector
    game_matrix_list: TxNxN array
    ----------
    Output:
    objective: negative log likelihood + squared l2 penalty
    '''
    # reshape beta into TxN array
    T, N = game_matrix_list.shape[0:2]
    beta = np.reshape(beta, [T, N])

    # compute l2 penalty
    l2_penalty = np.sum(np.square(beta[:-1] - beta[1:]))

    return model.neg_log_like(beta, game_matrix_list) + l_penalty * l2_penalty


def grad_l2_sq(beta, game_matrix_list, l):
    '''
    compute the gradient of the model (neg_log_like + l2_square)
    ----------
    Input:
    beta: TxN array or a TN vector
    game_matrix_list: TxNxN array
    ----------
    Output:
    objective: negative log likelihood + squared l2 penalty
    '''
    # reshape beta into TxN array
    T, N = game_matrix_list.shape[0:2]
    beta = np.reshape(beta, [T, N])

    # compute l2 penalty
    l2_grad = model.grad_nl(beta, game_matrix_list)
    l2_grad[N:] += l * 2 * ((beta[1:] - beta[:-1])).reshape(((T - 1) * N, 1))
    l2_grad[:-N] += l * 2 * ((beta[:-1] - beta[1:])).reshape(((T - 1) * N, 1))

    return l2_grad


def hess_l2_sq(beta, game_matrix_list, l):
    '''
    compute the Hessian of the model (neg_log_like + l2_square)
    ----------
    Input:
    beta: TxN array or a TN vector
    game_matrix_list: TxNxN array
    ----------
    Output:
    objective: negative log likelihood + squared l2 penalty
    '''
    # reshape beta into TxN array
    T, N = game_matrix_list.shape[0:2]
    beta = np.reshape(beta, [T, N])

    # compute l2 penalty
    l2_hess = model.hess_nl(beta, game_matrix_list)
    off_diag = np.array([2] + [0] * (N - 1) + [-1] + [0] * (N * (T - 1) - 1))
    l2_hess += l * 2 * sc.linalg.toeplitz(off_diag, off_diag)
    l2_hess[0:N, 0:N] -= l * 2 * np.diag(np.ones(N))
    l2_hess[-N:, -N:] -= l * 2 * np.diag(np.ones(N))
    return l2_hess


def prox_l2_sq(beta, s, l):
    '''
    proximal operator for l2-square-penalty
    '''
    n = np.array(beta).shape[0]

    # define banded matrix
    banded = np.block([
        [np.zeros([1, 1]), (-1) * 2 * s * l * np.ones([1, n - 1])],
        [(1 + 2 * s * l) * np.ones([1, 1]), (1 + 2 * 2 * s * l) *
         np.ones([1, n - 2]), (1 + 2 * s * l) * np.ones([1, 1])],
        [(-1) * 2 * s * l * np.ones([1, n - 1]), np.zeros([1, 1])]
    ])

    # solve banded @ beta* = beta
    return spl.solve_banded((1, 1), banded, beta, True, True, False)


def newton_l2_sq(data, l_penalty=1,
                 max_iter=1000, ths=1e-12,
                 step_init=1, max_back=200, a=0.01, b=0.3,
                 beta_init=None, verbose=False, out=sys.stdout):
    # initialize optimization
    T, N = data.shape[0:2]
    if beta_init is None:
        beta = np.zeros(data.shape[:2]).reshape((N * T, 1))
    else:
        beta = beta_init.reshape((N * T, 1))

    # initialize record
    objective_nt = [objective_l2_sq(beta, data, l_penalty)]
    if verbose:
        out.write("initial objective value: %f\n" % objective_nt[-1])
        out.flush()

    # iteration
    for i in range(max_iter):
        # compute gradient
        gradient = grad_l2_sq(beta, data, l_penalty)[1:]
        hessian = hess_l2_sq(beta, data, l_penalty)[1:, 1:]
        # backtracking
        obj_old = np.inf
        s = step_init
        beta_new = beta - 0  # make a copy

        for j in range(max_back):
            v = -sc.linalg.solve(hessian, gradient)
            beta_new[1:] = beta_new[1:] + s * v
            obj_new = objective_l2_sq(beta_new, data, l_penalty)

            if obj_new <= obj_old + b * s * gradient.T @ v:
                break
            s *= a

        beta = beta_new

        # objective value
        objective_nt.append(obj_new)
        obj_old = obj_new

        if verbose:
            out.write("%d-th Newton, objective value: %f\n" %
                      (i + 1, objective_nt[-1]))
            out.flush()
        if objective_nt[-2] - objective_nt[-1] < ths:
            if verbose:
                out.write("Converged!\n")
                out.flush()
            break
        elif i >= max_iter - 1:
            if verbose:
                out.write("Not converged.\n")
                out.flush()

    beta = beta.reshape((T, N))
    beta = beta - sum(beta[0, 0:N]) / N

    return objective_nt, beta


def pgd_l2_sq(data, l_penalty=1,
              max_iter=1000, ths=1e-12,
              step_init=0.5, max_back=200, a=0.2, b=0.5,
              beta_init=None, verbose=False, out=sys.stdout):
    # initialize optimization
    T, N = data.shape[0:2]
    if beta_init is None:
        beta = np.zeros(data.shape[:2])
    else:
        beta = beta_init
    nll = model.neg_log_like(beta, data)

    # initialize record
    objective_wback = [objective_l2_sq(beta, data, l_penalty)]
    if verbose:
        out.write("initial objective value: %f\n" % objective_wback[-1])
        out.flush()

    # iteration
    for i in range(max_iter):
        # compute gradient
        gradient = model.grad_nl(beta, data).reshape([T, N])

        # backtracking line search
        s = step_init
        for j in range(max_back):
            beta_new = prox_l2_sq(beta - s * gradient, s, l_penalty)
            beta_diff = beta_new - beta

            nll_new = model.neg_log_like(beta_new, data)
            nll_back = (nll + np.sum(gradient * beta_diff)
                        + np.sum(np.square(beta_diff)) / (2 * s))

            if nll_new <= nll_back:
                break
            s *= b

        # proximal gradient update
        beta = beta_new
        nll = nll_new

        # record objective value
        objective_wback.append(objective_l2_sq(beta, data, l_penalty))

        if verbose:
            out.write("%d-th PGD, objective value: %f\n" %
                      (i + 1, objective_wback[-1]))
            out.flush()
        if abs(objective_wback[-2] - objective_wback[-1]) < ths:
            if verbose:
                out.write("Converged!\n")
                out.flush()
            break
        elif i >= max_iter - 1:
            if verbose:
                out.write("Not converged.\n")
                out.flush()

    return objective_wback, beta

########################## l2 penalty ############################


def objective_l2(beta, game_matrix_list, l_penalty):
    '''
    compute the objective of the model (neg_log_like + l2)
    ----------
    Input:
    beta: TxN array or a TN vector
    game_matrix_list: TxNxN array
    l: coefficient of penalty term
    ----------
    Output:
    objective: negative log likelihood + l2 penalty
    '''
    # reshape beta into TxN array
    T, N = game_matrix_list.shape[0:2]
    beta = np.reshape(beta, [T, N])

    # compute l2 penalty
    diff = beta[:-1] - beta[1:]
    l2_penalty = sum([np.linalg.norm(diff[i]) for i in range(T - 1)])

    return model.neg_log_like(beta, game_matrix_list) + l_penalty * l2_penalty


def grad_l2(beta, game_matrix_list, l):
    '''
    compute the gradient of the model (neg_log_like + l2)
    ----------
    Input:
    beta: TxN array or a TN vector
    game_matrix_list: TxNxN array
    l: coefficient of penalty term
    ----------
    Output:
    objective: negative log likelihood + l2 penalty
    '''
    # reshape beta into TxN array
    T, N = game_matrix_list.shape[0:2]
    beta = np.reshape(beta, [T, N])

    # compute l2 penalty
    l2_grad = model.grad_nl(beta, game_matrix_list)
    diff = beta[1:] - beta[:-1]
    w = np.array([np.linalg.norm(diff[i]) for i in range(T - 1)])
    w[w != 0] = 1 / w[w != 0]
    l2_grad[N:] += l * np.array([diff[i] * w[i]
                                 for i in range(T - 1)]).reshape(((T - 1) * N, 1))

    diff = beta[:-1] - beta[1:]
    w = np.array([np.linalg.norm(diff[i]) for i in range(T - 1)])
    w[w != 0] = 1 / w[w != 0]
    l2_grad[:-N] += l * np.array([diff[i] * w[i]
                                  for i in range(T - 1)]).reshape(((T - 1) * N, 1))

    return l2_grad


def gd_l2(data, l_penalty=1,
          max_iter=1000, ths=1e-12,
          step_init=1, max_back=200, a=0.01, b=0.3,
          beta_init=None, verbose=False, out=sys.stdout):
    # initialize optimization
    T, N = data.shape[0:2]
    if beta_init is None:
        beta = np.zeros(data.shape[:2])
    else:
        beta = beta_init

    # initialize record
    objective_gd = [objective_l2(beta, data, l_penalty)]
    if verbose:
        out.write("initial objective value: %f\n" % objective_gd[-1])
        out.flush()

        # iteration
    for i in range(max_iter):
        # compute gradient
        gradient = grad_l2(beta, data, l_penalty).reshape([T, N])

        # backtracking
        s = step_init
        for j in range(max_back):
            # gradient update
            beta_new = beta - s * gradient
            obj_new = objective_l2(beta_new, data, l_penalty)
            if obj_new < objective_gd[-1] - b * \
                    s * np.sum(np.square(gradient)):
                break
            s *= a

        # objective value
        beta = beta_new
        objective_gd.append(obj_new)

        out.write("%d-th GD, objective value: %f\n" %
                  (i + 1, objective_gd[-1]))
        out.flush()
        if abs(objective_gd[-2] - objective_gd[-1]) < ths:
            if verbose:
                out.write("Converged!\n")
                out.flush()
            break
        elif i >= max_iter - 1:
            if verbose:
                out.write("Not converged.\n")
                out.flush()

    beta = beta.reshape((T, N))
    beta = beta - sum(beta[0, 0:N]) / N

    return objective_gd, beta


############## functions for ADMM ################

def obj_amlag(beta, data, A, muk, eta, thetak):
    return model.neg_log_like(beta, data) + (A @ beta).T @ muk + \
        eta / 2 * np.linalg.norm(A @ beta - thetak) ** 2


def admm_sub_beta(
        data,
        T,
        N,
        A,
        lam,
        eta,
        beta,
        muk,
        thetak,
        paras,
        out=sys.stdout):
    step_init, ths, max_iter, max_back, a, b = paras
    obj_old = np.inf

    for i in range(max_iter):
        # compute gradient
        gradient = model.grad_nl(beta, data) + A.T @ muk + \
            eta * A.T @ (A @ beta - thetak)
        hessian = model.hess_nl(beta, data) + eta * A.T @ A
        gradient = gradient[1:]
        hessian = hessian[1:, 1:]
        # proximal gradient update
        s = step_init
        beta_new = beta - 0  # make a copy

        for j in range(max_back):
            v = -sc.linalg.solve(hessian, gradient)
            beta_new[1:] = beta[1:] + s * v
            obj_new = obj_amlag(beta_new, data, A, muk, eta, thetak)

            if obj_new <= obj_old + b * s * gradient.T @ v:
                break
            s *= a

        beta = beta_new
        if abs(obj_old - obj_new) < ths:
            break
        elif i >= max_iter - 1:
            out.write("Not converged.\n")
            out.flush()
        obj_old = obj_new

    return beta


def prox_l2(t, x):
    if np.linalg.norm(x, 2) <= t:
        return 0
    else:
        return (1 - t / np.linalg.norm(x, 2)) * x


def prox_l1(t, x):
    return np.sign(x) * np.max(np.array([x * 0, np.abs(x) - t]), axis=0)


def admm_l2(data, l_penalty=1,
            max_iter=1000, ths=1e-12, eta=None,
            step_init=1, max_back=200, a=0.01, b=0.3,
            beta_init=None, verbose=False, out=sys.stdout, return_b_obj=False):
    # initialize optimization
    T, N = data.shape[0:2]
    if beta_init is None:
        beta = np.zeros(data.shape[:2]).reshape((N * T, 1))
    else:
        beta = beta_init.reshape((N * T, 1))

    # optimization parameters
    paras = [step_init, ths, max_iter, max_back, a, b]

    A = np.zeros(((T - 1) * N, T * N))
    for i in range(N):
        for t in range(T - 1):
            A[t * N + i, (t + 1) * N + i] = 1
            A[t * N + i, (t) * N + i] = -1

    if eta is None:
        eta = 20 * l_penalty

    thetak = np.zeros(((T - 1) * N, 1))
    muk = A @ beta - thetak

    # initialize record
    objective_admm_b_l2 = [objective_l2(beta, data, l_penalty)]
    objective_admm = [model.neg_log_like(
        beta, data) + l_penalty * np.linalg.norm(thetak)]
    if verbose:
        out.write("initial objective value: %f\n" % objective_admm[-1])
        out.flush()

    # iteration
    for i in range(max_iter):
        # compute gradient
        beta = admm_sub_beta(data, T, N, A, l_penalty, eta,
                             beta, muk, thetak, paras, out)
        thetak = prox_l2(l_penalty / eta, A @ beta + muk / eta)
        muk = muk + eta * (A @ beta - thetak)
        # objective value
        objective_admm_b_l2.append(objective_l2(beta, data, l_penalty))
        objective_admm.append(model.neg_log_like(
            beta, data) + l_penalty * np.linalg.norm(thetak))

        # if verbose
        #     print("%d-th ADMM, objective value: %f"%(i+1, objective_admm_b_l2[-1]))
        # if abs(objective_admm_b_l2[-2] - objective_admm_b_l2[-1]) < ths:
        #     print("Converged!")
        #     break
        if verbose:
            out.write("%d-th ADMM, objective value: %f\n" %
                      (i + 1, objective_admm[-1]))
            out.flush()
        if objective_admm[-2] - objective_admm[-1] < ths:
            if verbose:
                out.write("Converged!\n")
                out.flush()
            break
        elif i >= max_iter - 1:
            if verbose:
                out.write("Not converged.")
                out.flush()

    beta = beta - sum(beta[0:N]) / N
    beta = beta.reshape((T, N))

    if return_b_obj:
        return (objective_admm, objective_admm_b_l2), beta
    return objective_admm, beta

########################## l1 penalty ############################


def objective_l1(beta, game_matrix_list, l_penalty):
    '''
    compute the objective of the model (neg_log_like + l1)
    ----------
    Input:
    beta: TxN array or a TN vector
    game_matrix_list: TxNxN array
    l: coefficient of penalty term
    ----------
    Output:
    objective: negative log likelihood + l1 penalty
    '''
    # reshape beta into TxN array
    T, N = game_matrix_list.shape[0:2]
    beta = np.reshape(beta, [T, N])

    # compute l2 penalty
    diff = beta[:-1] - beta[1:]
    l1_penalty = sum([np.linalg.norm(diff[i], 1) for i in range(T - 1)])

    return model.neg_log_like(beta, game_matrix_list) + l_penalty * l1_penalty


def grad_l1(beta, game_matrix_list, l):
    '''
    compute the gradient of the model (neg_log_like + l1)
    ----------
    Input:
    beta: TxN array or a TN vector
    game_matrix_list: TxNxN array
    l: coefficient of penalty term
    ----------
    Output:
    objective: negative log likelihood + l1 penalty
    '''
    # reshape beta into TxN array
    T, N = game_matrix_list.shape[0:2]
    beta = np.reshape(beta, [T, N])

    # compute l1 penalty
    l1_grad = model.grad_nl(beta, game_matrix_list)
    diff = beta[1:] - beta[:-1]
    l1_grad[N:] += l * np.array([np.sign(diff[i])
                                 for i in range(T - 1)]).reshape(((T - 1) * N, 1))
    l1_grad[:-N] += l * np.array([-np.sign(diff[i])
                                  for i in range(T - 1)]).reshape(((T - 1) * N, 1))

    return l1_grad


def gd_l1(data, l_penalty=1,
          max_iter=50000, ths=1e-12, step_size=0.05,
          beta_init=None, verbose=False, out=sys.stdout):
    # initialize optimization
    T, N = data.shape[0:2]
    if beta_init is None:
        beta = np.zeros(data.shape[:2])
    else:
        beta = beta_init

    # initialize record
    objective_gd = [objective_l1(beta, data, l_penalty)]
    if verbose:
        out.write("initial objective value: %f\n" % objective_gd[-1])
        out.flush()

    # iteration
    for i in range(max_iter):
        # compute gradient
        gradient = grad_l1(beta, data, l_penalty).reshape([T, N])

        # proximal gradient update
        beta = beta - step_size / round((i + 11) / 10) * gradient

        # objective value
        objective_gd.append(objective_l1(beta, data, l_penalty))

        if verbose:
            if (i + 1) % 100 == 0:
                out.write("%d-th GD, objective value: %f\n" %
                          (i + 1, objective_gd[-1]))
                out.flush()

        if abs(objective_gd[-2] - objective_gd[-1]) < ths:
            if verbose:
                out.write("Converged!\n")
                out.flush()
            break
        elif i >= max_iter - 1:
            if verbose:
                out.write("Not converged.\n")
                out.flush()

    beta = beta.reshape((T, N))
    beta = beta - sum(beta[0, 0:N]) / N

    return objective_gd, beta


def admm_l1(data, l_penalty=1,
            max_iter=1000, ths=1e-12, eta=None,
            step_init=1, max_back=200, a=0.01, b=0.3,
            beta_init=None, verbose=False, out=sys.stdout, return_b_obj=False):
    # initialize optimization
    T, N = data.shape[0:2]
    if beta_init is None:
        beta = np.zeros(data.shape[:2]).reshape((N * T, 1))
    else:
        beta = beta_init.reshape((N * T, 1))

    # optimization parameters
    paras = [step_init, ths, max_iter, max_back, a, b]

    A = np.zeros(((T - 1) * N, T * N))
    for i in range(N):
        for t in range(T - 1):
            A[t * N + i, (t + 1) * N + i] = 1
            A[t * N + i, (t) * N + i] = -1

    if eta is None:
        eta = 20 * l_penalty

    thetak = np.zeros(((T - 1) * N, 1))
    muk = A @ beta - thetak

    # initialize record
    objective_admm_b_l1 = [objective_l1(beta, data, l_penalty)]
    objective_admm = [model.neg_log_like(beta, data) + l_penalty
                      * np.linalg.norm(thetak, 1)]
    if verbose:
        out.write("initial objective value: %f\n" % objective_admm[-1])
        out.flush()

    # iteration
    for i in range(max_iter):
        # compute gradient
        beta = admm_sub_beta(data, T, N, A, l_penalty, eta,
                             beta, muk, thetak, paras, out)
        thetak = prox_l1(l_penalty / eta, A @ beta + muk / eta)
        muk = muk + eta * (A @ beta - thetak)
        # objective value
        objective_admm_b_l1.append(objective_l1(beta, data, l_penalty))
        objective_admm.append(model.neg_log_like(
            beta, data) + l_penalty * np.linalg.norm(thetak, 1))

        # if verbose:
        #     print("%d-th ADMM, objective value: %f"%(i+1, objective_admm_b_l1[-1]))
        # if abs(objective_admm_b_l1[-2] - objective_admm_b_l1[-1]) < ths:
        #     print("Converged!")
        #     break
        if verbose:
            out.write("%d-th ADMM, objective value: %f\n" %
                      (i + 1, objective_admm[-1]))
            out.flush()
        if objective_admm[-2] - objective_admm[-1] < ths:
            if verbose:
                out.write("Converged!\n")
                out.flush()
            break
        elif i >= max_iter - 1:
            if verbose:
                out.write("Not converged.\n")
                out.flush()

    beta = beta - sum(beta[0:N]) / N
    beta = beta.reshape((T, N))

    if return_b_obj:
        return (objective_admm, objective_admm_b_l1), beta
    return objective_admm, beta
